import logging


def set_stream_logger(name="pinthesky", level=logging.DEBUG, fmt_string=None):
    if fmt_string is None:
        fmt_string = "%(asctime)s %(name)s [%(levelname)s] %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(fmt_string)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


class NullHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        pass


logging.getLogger('pinthesky').addHandler(NullHandler())



import logging
from pinthesky import events, input, output, upload, session, combiner, set_stream_logger
from pinthesky.camera import CameraThread
import argparse
import sys
import signal


VERSION = "0.1.2"


def create_parser():
    parser = argparse.ArgumentParser(
        description="Simple camera stream that captures motion video events")
    parser.add_argument(
        "--version",
        help="displays the current version",
        type=bool)
    parser.add_argument(
        "--combine-dir",
        help="the directory to combine video, defaults to motion_videos",
        default="motion_videos")
    parser.add_argument(
        "--rotation",
        help="rotate the video, valid arguments [0, 90, 180, 270]",
        default=270,
        type=int)
    parser.add_argument(
        "--resolution",
        help="camera resolution, defaults 640x480",
        default="640x480")
    parser.add_argument(
        "--framerate",
        help="framerate of the camera, defaults to 20",
        type=int,
        default=20)
    parser.add_argument(
        "--buffer",
        help="buffer size in seconds, defaults to 15",
        type=int,
        default=15)
    parser.add_argument(
        "--sensitivity",
        help="sensitivity of the motion detection math, default 10",
        type=int,
        default=10)
    parser.add_argument(
        "--event-input",
        help="file representing external input, default input.json",
        default="input.json")
    parser.add_argument(
        "--event-output",
        help="file representing external output, default output.json",
        default="output.json")
    parser.add_argument(
        "--configure-input",
        help="file for configuration input, default config-input.json",
        default="config-input.json")
    parser.add_argument(
        "--configure-output",
        help="file for configuration output, default config-output.json",
        default="config-output.json")
    parser.add_argument(
        "--thing-name",
        help="the AWS IoT ThingName for use in upload",
        default=None,
        required=False)
    parser.add_argument(
        "--thing-cert",
        help="the AWS IoT certificate associated to the Thing",
        default=None,
        required=False)
    parser.add_argument(
        "--thing-key",
        help="the AWS IoT certificate pair associated to the Thing",
        default=None,
        required=False)
    parser.add_argument(
        "--ca-cert",
        help="the root CA certificate to authenticate the certificate",
        default=None,
        required=False)
    parser.add_argument(
        "--credentials-endpoint",
        help="the AWS IoT Credentials Provider endpoint",
        default=None,
        required=False)
    parser.add_argument(
        "--role-alias",
        help="the AWS IoT Role Alias to pull credentials",
        default=None,
        required=False)
    parser.add_argument(
        "--bucket-name",
        help="the S3 bucket to upload motion detection files",
        default=None,
        required=False)
    parser.add_argument(
        "--bucket-prefix",
        help="the prefix to upload the motion files to, default motion_videos",
        default="motion_videos")
    parser.add_argument(
        "--recording-window",
        help="the recording window for the camera relative to the host time," +
        " ie: '08-18' for a ten hour window, defaults to always recordding",
        required=False,
        default=None)
    return parser


def main():
    parsed = create_parser().parse_args(sys.argv[1:])
    if parsed.version:
        print(VERSION)
        exit(0)
    # TODO: make this externally configurable
    set_stream_logger("pinthesky", level=logging.INFO)
    event_thread = events.EventThread()
    notify_thread = input.INotifyThread(events=event_thread)
    event_output = output.Output(output_file=parsed.event_output)
    auth_session = session.Session(
        cacert_path=parsed.ca_cert,
        cert_path=parsed.thing_cert,
        key_path=parsed.thing_key,
        role_alias=parsed.role_alias,
        thing_name=parsed.thing_name,
        credentials_endpoint=parsed.credentials_endpoint)
    video_uploader = upload.S3Upload(
        events=event_thread,
        bucket_name=parsed.bucket_name,
        bucket_prefix=parsed.bucket_prefix,
        session=auth_session)
    camera_thread = CameraThread(
        events=event_thread,
        sensitivity=parsed.sensitivity,
        resolution=tuple(map(int, parsed.resolution.split('x'))),
        rotation=parsed.rotation,
        framerate=parsed.framerate,
        recording_window=parsed.recording_window)
    video_combiner = combiner.VideoCombiner(
        events=event_thread,
        combine_dir=parsed.combine_dir)
    event_input_handler = input.InputHandler(events=event_thread)
    event_thread.on(camera_thread)
    event_thread.on(video_combiner)
    event_thread.on(video_uploader)
    event_thread.on(event_output)
    event_thread.on(event_input_handler)
    event_thread.on(auth_session)
    notify_thread.notify_change(parsed.event_input)
    notify_thread.notify_change(parsed.configure_output)
    event_thread.start()
    camera_thread.start()
    notify_thread.start()

    def signal_handler(signum, frame):
        notify_thread.stop()
        camera_thread.stop()
        event_thread.stop()

    signal.signal(signalnum=signal.SIGINT, handler=signal_handler)
    camera_thread.join()


if __name__ == "__main__":
    main()




class Handler():
    """
    Base handler that provides all of the methods to implement
    """
    def on_combine_end(self, event):
        """
        Handle the event signaling for when video is combined.
        Event field is: `combine_dir`
        """
        pass

    def on_flush_end(self, event):
        """
        Handle the event signaling the camera flushing.
        Event fields are `start_time`
        """
        pass

    def on_motion_start(self, event):
        """
        Handle the event signaling when motion is detected.
        """
        pass

    def on_upload_end(self, event):
        """
        Handle the event signaling when upload to S3 has finished.
        Event fields are `start_time` and `upload` which is a doucument
        containing `bucket_name` and `bucket_key`.
        """
        pass

    def on_file_change(self, event):
        """
        Handle the event signaling when there is a file change.
        Event fields are `content` which contains the content of the file.
        """
        pass



from math import floor
import logging
import threading
import time
import queue

from functools import partial
from pinthesky.handler import Handler

logger = logging.getLogger(__name__)
event_names = [
    'motion_start',
    'flush_end',
    'combine_end',
    'upload_end',
    'file_change'
]


class EventThread(threading.Thread):
    """
    This thread wraps a queue to flush events sequentially. A Handler could be
    added, or more general anonymous functions.
    """
    def __init__(self):
        super().__init__(daemon=True)
        self.event_queue = queue.Queue()
        self.running = True
        self.handlers = {}

    def on(self, handler: Handler):
        for event_name in event_names:
            method_name = f'on_{event_name}'
            method = getattr(handler, method_name)
            self.on_event(event_name=event_name, handler=partial(method))

    def on_event(self, event_name, handler):
        if event_name not in self.handlers:
            self.handlers[event_name] = []
        self.handlers[event_name].append(handler)

    def fire_event(self, event_name, context={}):
        event_data = {
            'name': event_name,
            'timestamp': floor(time.time())
        }
        if event_name in self.handlers:
            self.event_queue.put(dict(context, **event_data))

    def run(self):
        logger.info('Starting the event handler thread')
        while self.running:
            message = self.event_queue.get()
            if message['name'] in self.handlers:
                for handler in self.handlers[message['name']]:
                    handler(message)
            self.event_queue.task_done()

    def stop(self):
        self.running = False



import boto3
import os
import logging

from pinthesky.handler import Handler

logger = logging.getLogger(__name__)


class S3Upload(Handler):
    """
    Handles the `combine_end` to flush the video content to a specific path by
    thing name. Note: if the session is not connected to a remote IoT Thing,
    then this handle does nothing.
    """
    def __init__(self, events, bucket_name, bucket_prefix, session):
        self.events = events
        self.bucket_name = bucket_name
        self.bucket_prefix = bucket_prefix
        self.session = session

    def on_combine_end(self, event):
        creds = self.session.login()
        if self.bucket_name is not None and creds is not None:
            video = os.path.basename(event['combine_video'])
            loc = f'{self.bucket_prefix}/{self.session.thing_name}/{video}'
            logger.debug(f"Uploading to s3://{self.bucket_name}/{loc}")
            session = boto3.Session(
                creds['accessKeyId'],
                creds['secretAccessKey'],
                creds['sessionToken'])
            try:
                s3 = session.client('s3')
                with open(event['combine_video'], 'rb') as f:
                    s3.upload_fileobj(f, self.bucket_name, loc)
                self.events.fire_event('upload_end', {
                    'start_time': event['start_time'],
                    'upload': {
                        'bucket_name': self.bucket_name,
                        'bucket_key': loc
                    }
                })
            except RuntimeError as e:
                logger.error(
                    f'Failed to upload to s3://{self.bucket_name}/{loc}: {e}')
            finally:
                # TODO: add a failure strategy / retry attempt here
                os.remove(event['combine_video'])
